#!/usr/bin/python

# Processlog.py: Create a graph from proclog output.
# Version: 0.1
# Release: 2005-08-23
# Copyright (C) 2005 Fabien Nioles
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Todo: lot of bugs fixing, cleanup, documentation, etc, etc, etc.

from sys import argv, stderr
stderr = open("/dev/null", "a")
import os
import re
import gdchart
import gd
# import Gnuplot
# import Numeric
from math import log10

gdchart.defaults = gdchart.option()

# An helper... pretty there is some way to do it better.
def _rev(seq):
	if (len(seq) > 1):
		return (seq[-1],)+_rev(seq[:-1])
	else:
		return seq[:]

def _invcol(cols):
	return tuple(map(lambda x: 0xffffff - x, cols))

# This class is just a container for option values.
class Chart:
	def __init__(self):
		# Get a copy of the default options
		self.options = gdchart.defaults.copy()
		self.option(format=gdchart.GDC_PNG)
		self.option(other_threshold=1)

	def rotate_colors(self, num, colors):
		col = colors * (num/len(colors))
		col = col + colors[:(num%len(colors))]
		return col

	def set_color_pastel(self,num):
		colors=()
		step=0xffffff/num
		for i in range(0,0xffffff,step):
			colors = colors + (i,)	
		self.option(set_color=colors)
		return colors

	def set_color_grad(self,num,step,mask=0xffffff):
		colors = ()
		step = max(0xff/step,1)
		step = step + (step << 8) + (step << 16)
		for i in range(0x000000,0xffffff&mask,step&mask):
			colors = colors + (i,)
		# colors = + _rev(colors[1:-1])
		colors = self.rotate_colors(num, colors)
		self.option(set_color=colors)
		return colors
	
	def set_color_rainbow(self, num):
		colors = (0xff0000,0xffff00,0x00ff00,
		          0x00ffff,0x0000ff,0xff00ff)
		colors = self.rotate_colors(num, colors)
		self.option(set_color=colors)
		return colors

	def set_color(self,num):
		return self.set_color_grad(num,num,0x00ffff)

	def option(self, **args):
		# Save option values in the object's dictionary.
		self.options.update(args)

	def draw(self):
		# Put options into effect.  Derived class implementations of draw()
		# should call this method before calling gdchart.chart().
		apply(gdchart.option, (), self.options)

class BlockChart(Chart):
	def __init__(self):
		Chart.__init__(self)
		self.option(bg_color=0xffffff)

	def sum_up(self,data):
		sum = ()
		latest = data[0]
		sum = (latest,)
		for array in data[1:]:
			array = map(lambda x,y: x+y, latest, array)
			latest = array
			sum = sum + (latest,)
		return sum

	def draw(self, size, file, labels, *data):
		data = self.sum_up(data)
		Chart.draw(self)
		args = (gdchart.GDC_AREA, size, file, labels) + data
		apply(gdchart.chart, args)

class ProcChart(BlockChart):
	"Not used currently."
	def set_color(self, num):
		pass
	def set_spam_color(self, spam, num):
		if spam > 0:
			spamcol = self.set_color_grad(spam,spam,0x00ff00)
			spamcol = tuple(map(lambda x: x + 0xff0000, spamcol))
			# spamcol = _rev(tuple(map(lambda x: 0xffffff - x, spamcol)))
			# spamcol = _rev(spamcol)
		else:
			spamcol = ()
		if num > spam:
			numcol = self.set_color_grad(num-spam+1,num-spam+1,0x00ffff)[1:]
			# numcol = tuple(map(lambda x: x / 2, _invcol(numcol)))
		else:
			numcol = ()
		self.option(set_color=spamcol+numcol)
		return spamcol+numcol;
	def draw(self, size, file, spam, labels, *data):
		self.set_spam_color(spam, len(data))
		BlockChart.draw(self, size, file, labels, *data)
			
class proclog:
	def __init__(self):
		self.data = {}
		self.deffont = gd.gdFontSmall
		self.columns = 3
		self.target = os.path.join(os.environ['HOME'],'public_html')
	
	def filter_name(self, name):
		if name[0] == '#':
			return None
		if name[0] == '*':
			return None
		if name[:8] == 'formail ':
			return None
		if name == 'sa/':
			return 'SPAM:spam/' #+ ':assassin/'
		if name == 'spam-bogofilter/':
			return 'SPAM' + ':bogofilter/'
		if name == 'virus/':
			return 'SPAM:spam/' #+ ':virus/'
		if name == 'spam/unsure/':
			return 'UNSURE'
		if name == 'spam/virus/':
			return 'SPAM:spam/'
		if name == 'spam/assassin/':
			return 'SPAM:spam/'
		if name[:5] == 'spam/':
			return 'SPAM' + ':' + name[5:]
		if name == '/var/mail/fabien':
			return '_inbox'
		return name
	def add_file(self, key, name):
		d = {}
		f = open(name)
		for line in f.xreadlines():
			line = line.strip()
			list = re.split('\s*', line, 2)
			folder = list[2].strip()
			value = (int(list[1]),
				float(list[0])/(1024*1024))
			folder = self.filter_name(folder)
			if not folder: continue
			if d.has_key(folder):
				d[folder] = (d[folder][0] + value[0],
					d[folder][1] + value[1])
			else:
				d[folder] = value
		self.data[key] = d

	def process_data(self):
		self.folders = []
		self.folders_size = []
		self.folders_num = []
		self.keys = self.data.keys()
		self.keys.sort()
		for key in self.keys:
			d = self.data[key]
			for folder in d.keys():
				if (folder not in self.folders):
					print >>stderr, "New folder %s" % folder
					self.folders.append(folder)
					self.folders_size.append([])
					self.folders_num.append([])
		self.folders.sort()
		for key in self.keys:
			d = self.data[key]
			for findex in range(len(self.folders)):
				folder = self.folders[findex]
				size = self.folders_size[findex]
				num = self.folders_num[findex]
				if (folder in d.keys()):
					vnum, vsize = d[folder]
					size.append(vsize)
					num.append(vnum)
				else:
					print >>stderr, "Default value for %s: %s" % (key, folder)
					size.append(0)
					num.append(0)
		lastkey = self.keys[-1]
		print >>stderr, self.data[lastkey]
		for findex in range(len(self.folders)):
			print >>stderr, "%s: %f/%d" %(self.folders[findex],self.folders_size[findex][-1], self.folders_num[findex][-1])

	def c2t(self, color):
		red = color >> 16
		green = (color >> 8) & 0xff
		blue = color & 0xff
		return red, green, blue

	def alloc_color(self, img, color):
		col = self.c2t(color)
		idx = img.colorExact(col)
		if (idx != -1): return idx
		idx = img.colorAllocate(col)
		if (idx != -1): return idx
		print "Warning: can't allocate color!"
		idx = img.colorClosest(col)
		if (idx != -1): return idx
		raise "Can't find a slot for color %06x!" % color

	def draw_charts(self):
		spams = len(filter(lambda l: l[:4] == 'SPAM' or l == 'UNSURE', self.folders))
		self.chart = ProcChart()
		self.chart.option(title='debit de courriel')
		self.chart.option(ytitle='Mo/semaine')
		args = ((500,500), os.path.join(self.target,'size.png'), spams, self.keys)
		args = args + tuple(self.folders_size)
		apply(self.chart.draw, args)
		self.chart.option(ytitle='courriel/semaine')
		args = ((500,500), os.path.join(self.target,'num.png'), spams, self.keys)
		args = args + tuple(self.folders_num)
		apply(self.chart.draw, args)

	def draw_legend(self):
		lines = len(self.folders)/self.columns
		if (len(self.folders)%self.columns != 0):
			lines = lines+1
		dx, dy = gd.fontstrsize(self.deffont, 'test')
		sx, sy = (600, (lines+1)*(dy+5))
		img = gd.image((sx,sy))
		white = self.alloc_color(img, 0xffffff)
		black = self.alloc_color(img, 0)
		img.fill((0,0), white)
		colors = self.chart.options['set_color']
		for idx in xrange(len(self.folders)):
			folder = self.folders[idx]
			color = self.alloc_color(img, colors[idx])
			(dx, dy) = gd.fontstrsize(self.deffont, folder)
			y = 5+(dy+5)*(idx % lines)
			x = 5+(sx/self.columns+5)*(idx/lines)
			img.rectangle((x,y), (x+dy, y+dy), black, color)
			img.string(self.deffont, (x+dy+5, y), folder, black)
		img.writePng(os.path.join(self.target,'legend.png'))

	def draw(self):
		self.process_data()
		self.draw_charts()
		self.draw_legend()

if (__name__ == '__main__'):
	dict = {}
	files = argv[1:]
	log = proclog()
	for file in files[-78:]:
		key = file[-14:-4]
		log.add_file(key, file)
	log.draw()
